/*
 * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
 * Copyright (C) 2000 Thomas Roessler <roessler@does-not-exist.org>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "mime.h"
#include "sort.h"
#include "mailbox.h"
#include "copy.h"
#include "mx.h"
#include "pager.h"
/* #include "mutt_crypt.h" */
#include "rfc2047.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef USE_IMAP
#include "imap.h"
#endif

#ifdef BUFFY_SIZE
#include "buffy.h"
#endif



#ifdef HAVE_PGP
#include "pgp.h"
#endif

#ifdef HAVE_GPGME
#include "crypt.h"
#endif

#ifdef HAVE_SMIME
#include "smime.h"
#endif



#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <utime.h>

extern char *ReleaseDate;

/* The folder the user last saved to.  Used by ci_save_message() */
static char LastSaveFolder[_POSIX_PATH_MAX] = "";



static int is_in_mdn_list (ADDRESS *addr, LIST *p)
{
  if (addr->mailbox)
  {
    for (;p; p = p->next)
      if (!mutt_strncasecmp (addr->mailbox, p->data, mutt_strlen (p->data)))
	return 1;
  }
  return 0;
}


/* Compare 2 addresses.  Thisactually should do a case-insensitive
   compare on the mailbox part (according to rfc-2298) but we ignore
   that because case sensitive mailboxes are not common.
   
   FIXME: There is a similar but static function in send.c: mutt_addrcmp.
 */
static int address_compare (ADDRESS *a, ADDRESS *b)
{
  if (!a->mailbox || !b->mailbox)
    return-1;
  return ascii_strcasecmp (a->mailbox, b->mailbox);
}

static void write_address (FILE *fp, const char *tag, const char *prefix,
                           ADDRESS *a, int qualify)
{
  const char *fqdn = mutt_fqdn (1);
  char buffer[STRING];
  ADDRESS *aa;

  aa = rfc822_cpy_adr (a);
  if (fqdn && qualify)
    rfc822_qualify (aa, fqdn);
  rfc2047_encode_adrlist (aa, tag);
  *buffer = 0;
  rfc822_write_address (buffer, sizeof (buffer), aa);
  rfc822_free_address (&aa);

  fprintf (fp, "%s: %s%s\n", tag, prefix, buffer);
}




/* Record that the MDN as already been sent. */
static void mark_mdn_done (HEADER *cur)
{
  /* Hmmm: How should this be done.  IMHO the New Message mark should
     be sufficient but it is not exactly what we want: IF one select
     "ignore for now" the message will be flaged as old but the MDN
     was not sent and the next time it gets displayed, there won't be
     an indication whether it should be sent or not.  Also automatic
     tools won't be able to detect that status. */
  mutt_set_flag (Context, cur, M_MDNSENT, 1);
}

/* Construct a Message Disposition Notiifcation and send it out. */
static void mutt_send_mdn (HEADER *cur, int confirmed, ADDRESS *dnt)
{
  char fnamebuf[LONG_STRING];
  FILE *fp;
  HEADER *msg;
  BODY *t;
  char current_date_buf[SHORT_STRING], *current_date;
  ADDRESS *addr;

  mutt_make_date (current_date_buf, sizeof (current_date_buf));
  current_date = current_date_buf;
  current_date += strncmp (current_date_buf, "Date: ",6)? 0:6;
  if (*current_date && current_date[strlen(current_date)-1] =='\n')
    current_date[strlen(current_date)-1] = 0;

  msg = mutt_new_header ();
  msg->env = mutt_new_envelope ();
  msg->env->to = dnt; /* Note that is not a copy! */
  msg->env->subject = safe_strdup ("Disposition notification");

  t = mutt_new_body ();
  msg->content = t;
  t->type = TYPEMULTIPART;
  t->subtype = safe_strdup ("report");
  t->encoding = ENC7BIT;
  t->use_disp = 0;
  t->disposition = DISPINLINE;
  mutt_generate_boundary (&t->parameter);
  mutt_set_parameter ("report-type", "disposition-notification",
                      &t->parameter);

  t->parts = mutt_new_body ();
  t = t->parts;
  t->type = TYPETEXT;
  t->subtype = safe_strdup ("plain");
  t->use_disp = 0;
  t->disposition = DISPINLINE;
  t->encoding = ENC7BIT;
  mutt_mktemp (fnamebuf);
  t->filename = safe_strdup (fnamebuf);
  t->unlink = 1; 
  if (!(fp = safe_fopen (t->filename, "w")))
  {
    mutt_perror (t->filename);
    /* fixme: cleanup */
    return;
  }
  fprintf (fp, "Original message info:\n\n");
  if (cur->env->date)
    fprintf (fp,       "    Date: %s\n", cur->env->date);
  if (cur->env->to)
    write_address (fp, "      To", "", cur->env->to, 0);
  if (cur->env->subject)
    fprintf (fp,       " Subject: %s\n", cur->env->subject);
  fprintf (fp, "\n"
    "The message has been displayed at %s.\n"
    "This is no guarantee that the message has been read or understood.\n\n",
     current_date);
  fclose (fp);


  t->next = mutt_new_body ();
  t = t->next;
  t->type = TYPEMESSAGE;
  t->subtype = safe_strdup ("disposition-notification");
  t->use_disp = 0;
  t->disposition = DISPINLINE;
  t->encoding = ENC7BIT;
  mutt_mktemp (fnamebuf);
  t->filename = safe_strdup (fnamebuf);
  t->unlink = 1; 
  if (!(fp = safe_fopen (t->filename, "w")))
  {
    mutt_perror (t->filename);
    /* fixme: cleanup */
    return;
  }

  fprintf (fp, "Reporting-UA: %s; Mutt/%s\n", Fqdn, MUTT_VERSION);
  /*fprintf (fp, "Original-Recipient: rfc822;%s\n");*/

  addr = mutt_default_from ();
  write_address (fp, "Final-Recipient", "rfc822;", addr, 1);
  rfc822_free_address (&addr);

  if (cur->env->message_id)
    fprintf (fp, "Original-Message-ID: %s\n", cur->env->message_id);

  fprintf (fp, "Disposition: manual-action/MDN-sent-%s; displayed\n",
           confirmed? "manually":"automatically");
  fclose (fp);

  if (!mutt_send_message_direct (msg, 1))
    mutt_message (_("Message Disposition Notification sent."));

  msg->env->to = NULL; /* we didn't used a copy */
  mutt_free_envelope (&msg->env); msg->env = NULL;

  mark_mdn_done (cur);
}


/* Have a look at the current message and decide whether a MDN should
   be sent.  Return true if the user selected to ignore it for now. */
int mutt_handle_mdn (HEADER *cur)
{
  int must_ask = 0;
  int ask_this = 0;
  int unknown_required = 0;
  int confirmed = 0;
  PARAMETER *parm;
  ADDRESS *dnt_addr = NULL;
  ADDRESS *a;

  if (!option (OPTMDNENABLE) )
    return 0; /* MDN support not enabled */

  /* Do we have a notification header? */
  if (!cur->env || !cur->env->dnt)
    return 0; /* Nothing to do. */
  if (cur->mdnsent)
    return 0; /* Already sent. */

  /* Check whether one of the notification addresses is in the deny
     list. Add all non-denied addresses to a new list. */
  for (a = cur->env->dnt; a; a = a->next)
  {
    if (!is_in_mdn_list (a, MdnDeny))
    {
      ADDRESS *aa = rfc822_cpy_adr_real (a);
      rfc822_append (&dnt_addr, aa);
      rfc822_free_address (&aa);
    }
  }

  if (!dnt_addr)
    {
      mark_mdn_done (cur);
      return 0; /* All addresses are in the deny list. */
    }

  /* Check whether we can handle all required options.  RFC2298 does
     not define any yet, but we must be prepared for future extensions
     and X-foo attributes. */
  for (parm=cur->env->dno; parm; parm = parm->next)
  {
    if (!ascii_strcasecmp (parm->value, "required"))
      unknown_required = 1;
    else if (ascii_strcasecmp (parm->value, "optional"))
      must_ask = 1; /* invalid importance value */
  }
  
  if (unknown_required)
  {
    /* We don't sent a failed messages at all. */
    mark_mdn_done (cur);
    mutt_message (_("Required Disposition Notification Option unknown "
                    "- DNT ignored"));
    rfc822_free_address (&dnt_addr);
    return 0;
  }


  /* Check that the Return-Path matches the DNT, a Return-Path exists
     and that it does not comtain more than one address.*/
  if (dnt_addr->next)
    must_ask = 1; /* More than one address. */
  else if (!cur->env->return_path)
    must_ask = 1;
  else if (cur->env->return_path->next)
    must_ask = 1;
  else if (address_compare (dnt_addr, cur->env->return_path))
    must_ask = 1; /* Addresses don't match. */
  else if ( option (OPTMDNCONFIRM) )
    ask_this = 1;

  /* Check the allow list, which overrides the above. */
  if (must_ask || ask_this)
  {
    for (a = dnt_addr; a; a = a->next)
      if (!is_in_mdn_list (a, MdnAllow))
        break;
    if (!a) /* All addresses are in the allow list. */
      must_ask = ask_this = 0;
  }

 
  /* Get confirmation if required. */
  if (must_ask)
  {
    switch (mutt_yesorno (_("Disposition-Notification request is questionable."
                            "  Really send?"), M_NO))
    {
      case M_YES:
      	break;
      case M_NO:
        mark_mdn_done (cur);
        rfc822_free_address (&dnt_addr);
        return 0;
      default:
        /* ^G */ 
        rfc822_free_address (&dnt_addr);
        return 1; /* ignore/postponed */
    }
    confirmed = 1;
  }
  else if (ask_this)
  {
    switch (mutt_yesorno (_("Disposition-Notification requested.  "
                            "Send?"), M_YES))
    {
      case M_YES:
      	break;
      case M_NO:
        mark_mdn_done (cur);
        rfc822_free_address (&dnt_addr);
        return 0;
      default:
        /* ^G */ 
        rfc822_free_address (&dnt_addr);
        return 1; /* ignore/postponed */
    }
    confirmed = 1;
  }

  mutt_send_mdn (cur, confirmed, dnt_addr);
  rfc822_free_address (&dnt_addr);
  return 0;
}



int mutt_display_message (HEADER *cur)
{
  char tempfile[_POSIX_PATH_MAX], buf[LONG_STRING];
  int rc = 0, builtin = 0;
  int cmflags = M_CM_DECODE | M_CM_DISPLAY | M_CM_CHARCONV;
  FILE *fpout = NULL;
  FILE *fpfilterout = NULL;
  pid_t filterpid = -1;
  int res;

  snprintf (buf, sizeof (buf), "%s/%s", TYPE (cur->content),
	    cur->content->subtype);

  mutt_parse_mime_message (Context, cur);
  mutt_message_hook (Context, cur, M_MESSAGEHOOK);

#if defined (HAVE_PGP) || defined (HAVE_SMIME) || defined(HAVE_GPGME)
  /* see if PGP is needed for this message.  if so, we should exit curses */
  if (cur->security)
  {
    if (cur->security & ENCRYPT)
    {
#ifdef HAVE_SMIME
      if (cur->security & APPLICATION_SMIME)
	smime_getkeys (cur->env);
#endif
      if(!crypt_valid_passphrase(cur->security))
	return 0;

      cmflags |= M_CM_VERIFY;
    }
    else if (cur->security & SIGN)
    {
      /* find out whether or not the verify signature */
      if (query_quadoption (OPT_VERIFYSIG, _("Verify PGP signature?")) == M_YES)
      {
	cmflags |= M_CM_VERIFY;
      }
    }
  }
  
  if (cmflags & M_CM_VERIFY || cur->security & ENCRYPT)
  {
#ifdef HAVE_PGP
    if (cur->security & APPLICATION_PGP)
    {
      if (cur->env->from)
        pgp_invoke_getkeys (cur->env->from);

      mutt_message _("Invoking PGP...");
    }
#endif

#ifdef HAVE_SMIME
    if (cur->security & APPLICATION_SMIME)
      mutt_message _("Invoking OpenSSL...");
#endif
#ifdef HAVE_GPGME
    if (cur->security & (APPLICATION_PGP|APPLICATION_SMIME))
    {
#if 0 /* FIXME */
      if (cur->env->from)
        pgp_invoke_getkeys (cur->env->from);
#endif

      mutt_message _("Invoking GnuPG...");
    }
#endif
  }
#endif /*  defined (HAVE_PGP) || defined (HAVE_SMIME) */


  mutt_mktemp (tempfile);
  if ((fpout = safe_fopen (tempfile, "w")) == NULL)
  {
    mutt_error _("Could not create temporary file!");
    return (0);
  }

  if (DisplayFilter && *DisplayFilter) 
  {
    fpfilterout = fpout;
    fpout = NULL;
    /* mutt_endwin (NULL); */
    filterpid = mutt_create_filter_fd (DisplayFilter, &fpout, NULL, NULL,
				       -1, fileno(fpfilterout), -1);
    if (filterpid < 0)
    {
      mutt_error (_("Cannot create display filter"));
      safe_fclose (&fpfilterout);
      unlink (tempfile);
      return 0;
    }
  }

  if (!Pager || mutt_strcmp (Pager, "builtin") == 0)
    builtin = 1;
  else
  {
    mutt_make_string (buf, sizeof (buf), NONULL(PagerFmt), Context, cur);
    fputs (buf, fpout);
    fputs ("\n\n", fpout);
  }

  res = mutt_copy_message (fpout, Context, cur, cmflags,
       	(option (OPTWEED) ? (CH_WEED | CH_REORDER) : 0) | CH_DECODE | CH_FROM);
  if ((safe_fclose (&fpout) != 0 && errno != EPIPE) || res == -1)
  {
    mutt_error (_("Could not copy message"));
    if (fpfilterout != NULL)
    {
      mutt_wait_filter (filterpid);
      safe_fclose (&fpfilterout);
    }
    mutt_unlink (tempfile);
    return 0;
  }

  mutt_handle_mdn (cur);

  if (fpfilterout != NULL && mutt_wait_filter (filterpid) != 0)
    mutt_any_key_to_continue (NULL);

  safe_fclose (&fpfilterout);	/* XXX - check result? */

  
#if defined(HAVE_PGP) || defined(HAVE_SMIME) || defined(HAVE_GPGME)
  /* update crypto information for this message */
  cur->security |= crypt_query (cur->content);
  
  /* Remove color cache for this message, in case there
     are color patterns for both ~g and ~V */
  cur->pair = 0;
#endif

  if (builtin)
  {
    pager_t info;

#if defined(HAVE_SMIME) || defined(HAVE_GPGME)
    if (cur->security & APPLICATION_SMIME && (cmflags & M_CM_VERIFY))
    {
      if (cur->security & GOODSIGN)
      {
#ifdef HAVE_GPGME  /* FIXME: Implement verify sender - but why here?? */
	  mutt_message ( _("S/MIME signature successfully verified."));
#else
	if (!smime_verify_sender(cur))
	  mutt_message ( _("S/MIME signature successfully verified."));
	else
	  mutt_error ( _("S/MIME certificate owner does not match sender."));
#endif
      }
      else if (cur->security & SIGN || cur->security & BADSIGN)
	mutt_error ( _("S/MIME signature could NOT be verified."));
    }
#endif

#if defined(HAVE_PGP) || defined(HAVE_GPGME)
    if ((cur->security & APPLICATION_PGP) && (cmflags & M_CM_VERIFY))
      mutt_message ((cur->security & GOODSIGN) ?
		    _("PGP signature successfully verified.") :
		    _("PGP signature could NOT be verified."));
#endif

    /* Invoke the builtin pager */
    memset (&info, 0, sizeof (pager_t));
    info.hdr = cur;
    info.ctx = Context;
    rc = mutt_pager (NULL, tempfile, M_PAGER_MESSAGE, &info);
  }
  else
  {
    int r;

    mutt_endwin (NULL);
    snprintf (buf, sizeof (buf), "%s %s", NONULL(Pager), tempfile);
    if ((r = mutt_system (buf)) == -1)
      mutt_error (_("Error running \"%s\"!"), buf);
    unlink (tempfile);
    keypad (stdscr, TRUE);
    if (r != -1)
      mutt_set_flag (Context, cur, M_READ, 1);
    if (r != -1 && option (OPTPROMPTAFTER))
    {
      mutt_ungetch (mutt_any_key_to_continue _("Command: "), 0);
      rc = km_dokey (MENU_PAGER);
    }
    else
      rc = 0;
  }

  return rc;
}

void ci_bounce_message (HEADER *h, int *redraw)
{
  char prompt[SHORT_STRING];
  char buf[HUGE_STRING] = { 0 };
  ADDRESS *adr = NULL;
  int rc;

  if(h)
    strfcpy(prompt, _("Bounce message to: "), sizeof(prompt));
  else
    strfcpy(prompt, _("Bounce tagged messages to: "), sizeof(prompt));
  
  rc = mutt_get_field (prompt, buf, sizeof (buf), M_ALIAS);

  if (option (OPTNEEDREDRAW))
  {
    unset_option (OPTNEEDREDRAW);
    *redraw = REDRAW_FULL;
  }

  if (rc || !buf[0])
    return;

  if (!(adr = rfc822_parse_adrlist (adr, buf)))
  {
    mutt_error _("Error parsing address!");
    return;
  }

  adr = mutt_expand_aliases (adr);

  buf[0] = 0;
  rfc822_write_address (buf, sizeof (buf), adr);

#define extra_space (15 + 7 + 2)
  /*
   * This is the printing width of "...? ([y=yes]/n=no): ?" plus 2
   * for good measure. This is not ideal. FIXME.
   */
  snprintf (prompt, sizeof (prompt) - 4,
           (h ? _("Bounce message to %s") : _("Bounce messages to %s")), buf);
  mutt_format_string (prompt, sizeof (prompt) - 4,
		      0, COLS-extra_space, 0, 0,
		      prompt, sizeof (prompt), 0);
  strcat (prompt, "...?");	/* __STRCAT_CHECKED__ */
  if (mutt_yesorno (prompt, 1) != 1)
  {
    rfc822_free_address (&adr);
    CLEARLINE (LINES-1);
    return;
  }

  mutt_bounce_message (NULL, h, adr);
  rfc822_free_address (&adr);
  mutt_message (h ? _("Message bounced.") : _("Messages bounced."));
}

static void pipe_set_flags (int decode, int print, int *cmflags, int *chflags)
{
  if (decode)
  {
    *cmflags |= M_CM_DECODE | M_CM_CHARCONV;
    *chflags |= CH_DECODE | CH_REORDER;
    
    if (option (OPTWEED))
    {
      *chflags |= CH_WEED;
      *cmflags |= M_CM_WEED;
    }
  }
  
  if (print)
    *cmflags |= M_CM_PRINTING;
  
}

void pipe_msg (HEADER *h, FILE *fp, int decode, int print)
{
  int cmflags = 0;
  int chflags = CH_FROM;
  
  pipe_set_flags (decode, print, &cmflags, &chflags);

#if defined(HAVE_PGP) || defined(HAVE_SMIME) || defined(HAVE_GPGME)
    if (decode && h->security & ENCRYPT)
    {
      if(!crypt_valid_passphrase(h->security))
	return;
      endwin ();
    }
#endif


  if (decode)
    mutt_parse_mime_message (Context, h);

  mutt_copy_message (fp, Context, h, cmflags, chflags);
}


/* the following code is shared between printing and piping */

static int _mutt_pipe_message (HEADER *h, char *cmd,
			       int decode,
			       int print,
			       int split,
			       char *sep)
{
  
  int i, rc = 0;
  pid_t thepid;
  FILE *fpout;
  
/*   mutt_endwin (NULL); 

     is this really needed here ? 
     it makes the screen flicker on pgp and s/mime messages,
     before asking for a passphrase...
                                     Oliver Ehli */
  if (h)
  {

    mutt_message_hook (Context, h, M_MESSAGEHOOK);

#if defined(HAVE_PGP) || defined(HAVE_SMIME) || defined(HAVE_GPGME)
    if (decode)
    {
      mutt_parse_mime_message (Context, h);
      if(h->security & ENCRYPT && !crypt_valid_passphrase(h->security))
	return 1;
    }
    mutt_endwin (NULL);
#endif

    if ((thepid = mutt_create_filter (cmd, &fpout, NULL, NULL)) < 0)
    {
      mutt_perror _("Can't create filter process");
      return 1;
    }
      
    pipe_msg (h, fpout, decode, print);
    fclose (fpout);
    rc = mutt_wait_filter (thepid);
  }
  else
  { /* handle tagged messages */



#if defined(HAVE_PGP) || defined(HAVE_SMIME) || defined(HAVE_GPGME)

    if (decode)
    {
      for (i = 0; i < Context->vcount; i++)
	if(Context->hdrs[Context->v2r[i]]->tagged)
	{
	  mutt_message_hook (Context, Context->hdrs[Context->v2r[i]], M_MESSAGEHOOK);
	  mutt_parse_mime_message(Context, Context->hdrs[Context->v2r[i]]);
	  if (Context->hdrs[Context->v2r[i]]->security & ENCRYPT &&
	      !crypt_valid_passphrase(Context->hdrs[Context->v2r[i]]->security))
	    return 1;
	}
    }
#endif
    
    if (split)
    {
      for (i = 0; i < Context->vcount; i++)
      {
        if (Context->hdrs[Context->v2r[i]]->tagged)
        {
	  mutt_message_hook (Context, Context->hdrs[Context->v2r[i]], M_MESSAGEHOOK);
	  mutt_endwin (NULL);
	  if ((thepid = mutt_create_filter (cmd, &fpout, NULL, NULL)) < 0)
	  {
	    mutt_perror _("Can't create filter process");
	    return 1;
	  }
          pipe_msg (Context->hdrs[Context->v2r[i]], fpout, decode, print);
          /* add the message separator */
          if (sep)  fputs (sep, fpout);
	  safe_fclose (&fpout);
	  if (mutt_wait_filter (thepid) != 0)
	    rc = 1;
        }
      }
    }
    else
    {
      mutt_endwin (NULL);
      if ((thepid = mutt_create_filter (cmd, &fpout, NULL, NULL)) < 0)
      {
	mutt_perror _("Can't create filter process");
	return 1;
      }
      for (i = 0; i < Context->vcount; i++)
      {
        if (Context->hdrs[Context->v2r[i]]->tagged)
        {
	  mutt_message_hook (Context, Context->hdrs[Context->v2r[i]], M_MESSAGEHOOK);
          pipe_msg (Context->hdrs[Context->v2r[i]], fpout, decode, print);
          /* add the message separator */
          if (sep) fputs (sep, fpout);
        }
      }
      safe_fclose (&fpout);
      if (mutt_wait_filter (thepid) != 0)
	rc = 1;
    }
  }

  if (rc || option (OPTWAITKEY))
    mutt_any_key_to_continue (NULL);
  return rc;
}

void mutt_pipe_message (HEADER *h)
{
  char buffer[LONG_STRING];

  buffer[0] = 0;
  if (mutt_get_field (_("Pipe to command: "), buffer, sizeof (buffer), M_CMD)
      != 0 || !buffer[0])
    return;

  mutt_expand_path (buffer, sizeof (buffer));
  _mutt_pipe_message (h, buffer,
		      option (OPTPIPEDECODE),
		      0, 
		      option (OPTPIPESPLIT),
		      PipeSep);
}

void mutt_print_message (HEADER *h)
{

  if (quadoption (OPT_PRINT) && (!PrintCmd || !*PrintCmd))
  {
    mutt_message (_("No printing command has been defined."));
    return;
  }
  
  if (query_quadoption (OPT_PRINT,
			h ? _("Print message?") : _("Print tagged messages?"))
		  	!= M_YES)
    return;

  if (_mutt_pipe_message (h, PrintCmd,
			  option (OPTPRINTDECODE),
			  1,
			  option (OPTPRINTSPLIT),
			  "\f") == 0)
    mutt_message (h ? _("Message printed") : _("Messages printed"));
  else
    mutt_message (h ? _("Message could not be printed") :
		  _("Messages could not be printed"));
}


int mutt_select_sort (int reverse)
{
  int method = Sort; /* save the current method in case of abort */

  switch (mutt_multi_choice (reverse ?
			     _("Rev-Sort (d)ate/(f)rm/(r)ecv/(s)ubj/t(o)/(t)hread/(u)nsort/si(z)e/s(c)ore?: ") :
			     _("Sort (d)ate/(f)rm/(r)ecv/(s)ubj/t(o)/(t)hread/(u)nsort/si(z)e/s(c)ore?: "),
			     _("dfrsotuzc")))
  {
  case -1: /* abort - don't resort */
    return -1;

  case 1: /* (d)ate */
    Sort = SORT_DATE;
    break;

  case 2: /* (f)rm */
    Sort = SORT_FROM;
    break;
  
  case 3: /* (r)ecv */
    Sort = SORT_RECEIVED;
    break;
  
  case 4: /* (s)ubj */
    Sort = SORT_SUBJECT;
    break;
  
  case 5: /* t(o) */
    Sort = SORT_TO;
    break;
  
  case 6: /* (t)hread */
    Sort = SORT_THREADS;
    break;
  
  case 7: /* (u)nsort */
    Sort = SORT_ORDER;
    break;
  
  case 8: /* si(z)e */
    Sort = SORT_SIZE;
    break;
  
  case 9: /* s(c)ore */ 
    Sort = SORT_SCORE;
    break;
  }
  if (reverse)
    Sort |= SORT_REVERSE;

  return (Sort != method ? 0 : -1); /* no need to resort if it's the same */
}

/* invoke a command in a subshell */
void mutt_shell_escape (void)
{
  char buf[LONG_STRING];

  buf[0] = 0;
  if (mutt_get_field (_("Shell command: "), buf, sizeof (buf), M_CMD) == 0)
  {
    if (!buf[0] && Shell)
      strfcpy (buf, Shell, sizeof (buf));
    if(buf[0])
    {
      CLEARLINE (LINES-1);
      mutt_endwin (NULL);
      fflush (stdout);
      if (mutt_system (buf) != 0 || option (OPTWAITKEY))
	mutt_any_key_to_continue (NULL);
    }
  }
}

/* enter a mutt command */
void mutt_enter_command (void)
{
  BUFFER err, token;
  char buffer[LONG_STRING], errbuf[SHORT_STRING];
  int r;

  buffer[0] = 0;
  if (mutt_get_field (":", buffer, sizeof (buffer), M_COMMAND) != 0 || !buffer[0])
    return;
  err.data = errbuf;
  err.dsize = sizeof (errbuf);
  memset (&token, 0, sizeof (token));
  r = mutt_parse_rc_line (buffer, &token, &err);
  FREE (&token.data);
  if (errbuf[0])
  {
    /* since errbuf could potentially contain printf() sequences in it,
       we must call mutt_error() in this fashion so that vsprintf()
       doesn't expect more arguments that we passed */
    if (r == 0)
      mutt_message ("%s", errbuf);
    else
      mutt_error ("%s", errbuf);
  }
}

void mutt_display_address (ENVELOPE *env)
{
  char *pfx = NULL;
  char buf[SHORT_STRING];
  ADDRESS *adr = NULL;

  adr = mutt_get_address (env, &pfx);

  if (!adr) return;

  buf[0] = 0;
  rfc822_write_address (buf, sizeof (buf), adr);
  mutt_message ("%s: %s", pfx, buf);
}

static void set_copy_flags (HEADER *hdr, int decode, int decrypt, int *cmflags, int *chflags)
{
  *cmflags = 0;
  *chflags = CH_UPDATE_LEN;
  
#if defined(HAVE_PGP) || defined(HAVE_SMIME) || defined(HAVE_GPGME)
  if (!decode && decrypt && (hdr->security & ENCRYPT))
  {
#if defined(HAVE_PGP) || defined(HAVE_GPGME)
    if (mutt_is_multipart_encrypted(hdr->content))
    {
      *chflags = CH_NONEWLINE | CH_XMIT | CH_MIME;
      *cmflags = M_CM_DECODE_PGP;
    }
    else if (mutt_is_application_pgp (hdr->content) & ENCRYPT)
      decode = 1;
#endif
#if (defined(HAVE_PGP) && defined(HAVE_SMIME)) || defined(HAVE_GPGME)
    else
#endif
#if defined(HAVE_SMIME) || defined(HAVE_GPGME)
	  if (mutt_is_application_smime(hdr->content) & ENCRYPT)
    {
      *chflags = CH_NONEWLINE | CH_XMIT | CH_MIME;
      *cmflags = M_CM_DECODE_SMIME;
    }
#endif
  }
#endif /* defined(HAVE_PGP) || defined(HAVE_SMIME) */

  if (decode)
  {
    *chflags = CH_XMIT | CH_MIME | CH_TXTPLAIN;
    *cmflags = M_CM_DECODE | M_CM_CHARCONV;
  }

  /* respect $weed only if decode doesn't kick in
   * for decrypt.
   */

  if (decode && !decrypt && option (OPTWEED))
  {
    *chflags |= CH_WEED;
    *cmflags |= M_CM_WEED;
  }
}

void _mutt_save_message (HEADER *h, CONTEXT *ctx, int delete, int decode, int decrypt)
{
  int cmflags, chflags;
  
  set_copy_flags (h, decode, decrypt, &cmflags, &chflags);

  if (decode || decrypt)
    mutt_parse_mime_message (Context, h);

  if (mutt_append_message (ctx, Context, h, cmflags, chflags) == 0 && delete)
  {
    mutt_set_flag (Context, h, M_DELETE, 1);
    if (option (OPTDELETEUNTAG))
      mutt_set_flag (Context, h, M_TAG, 0);
  }
}

/* returns 0 if the copy/save was successful, or -1 on error/abort */
int mutt_save_message (HEADER *h, int delete, 
		       int decode, int decrypt, int *redraw)
{
  int i, need_buffy_cleanup;
#if defined(HAVE_PGP) || defined(HAVE_SMIME) || defined(HAVE_GPGME)
  int need_passphrase = 0, app=0;
#endif
  char prompt[SHORT_STRING], buf[_POSIX_PATH_MAX];
  CONTEXT ctx;
  struct stat st;
#ifdef BUFFY_SIZE
  BUFFY *tmp = NULL;
#else
  struct utimbuf ut;
#endif

  *redraw = 0;

  
  snprintf (prompt, sizeof (prompt),
	    decode  ? (delete ? _("Decode-save%s to mailbox") :
		       _("Decode-copy%s to mailbox")) :
	    (decrypt ? (delete ? _("Decrypt-save%s to mailbox") :
			_("Decrypt-copy%s to mailbox")) :
	     (delete ? _("Save%s to mailbox") : _("Copy%s to mailbox"))),
	    h ? "" : _(" tagged"));
  

  if (h)
  {
#if defined(HAVE_PGP) || defined(HAVE_SMIME) || defined(HAVE_GPGME)
      need_passphrase = h->security & ENCRYPT;
      app = h->security;
#endif
    mutt_message_hook (Context, h, M_MESSAGEHOOK);
    mutt_default_save (buf, sizeof (buf), h);
  }
  else
  {
    /* look for the first tagged message */

    for (i = 0; i < Context->vcount; i++)
    {
      if (Context->hdrs[Context->v2r[i]]->tagged)
      {
	h = Context->hdrs[Context->v2r[i]];
	break;
      }
    }


    if (h)
    {
      mutt_message_hook (Context, h, M_MESSAGEHOOK);
      mutt_default_save (buf, sizeof (buf), h);
#if defined(HAVE_PGP) || defined(HAVE_SMIME) || defined(HAVE_GPGME)
      need_passphrase = h->security & ENCRYPT;
      app = h->security;
#endif
      h = NULL;
    }
  }

  mutt_pretty_mailbox (buf);
  if (mutt_enter_fname (prompt, buf, sizeof (buf), redraw, 0) == -1)
    return (-1);

  if (*redraw != REDRAW_FULL)
  {
    if (!h)
      *redraw = REDRAW_INDEX | REDRAW_STATUS;
    else
      *redraw = REDRAW_STATUS;
  }

  if (!buf[0])
    return (-1);
 
  /* This is an undocumented feature of ELM pointed out to me by Felix von
   * Leitner <leitner@prz.fu-berlin.de>
   */
  if (mutt_strcmp (buf, ".") == 0)
    strfcpy (buf, LastSaveFolder, sizeof (buf));
  else
    strfcpy (LastSaveFolder, buf, sizeof (LastSaveFolder));

  mutt_expand_path (buf, sizeof (buf));

  /* check to make sure that this file is really the one the user wants */
  if (!mutt_save_confirm (buf, &st))
    return -1;

  if (need_passphrase && (decode || decrypt) && !crypt_valid_passphrase(app))
    return -1;


  mutt_message (_("Copying to %s..."), buf);
  
#ifdef USE_IMAP
  if (Context->magic == M_IMAP && 
      !(decode || decrypt) && mx_is_imap (buf))
  {
    switch (imap_copy_messages (Context, h, buf, delete))
    {
      /* success */
      case 0: mutt_clear_error (); return 0;
      /* non-fatal error: fall through to fetch/append */
      case 1: break;
      /* fatal error, abort */
      case -1: return -1;
    }
  }
#endif

  if (mx_open_mailbox (buf, M_APPEND, &ctx) != NULL)
  {
    if (h)
      _mutt_save_message(h, &ctx, delete, decode, decrypt);
    else
    {
      for (i = 0; i < Context->vcount; i++)
      {
	if (Context->hdrs[Context->v2r[i]]->tagged)
	{
	  mutt_message_hook (Context, Context->hdrs[Context->v2r[i]], M_MESSAGEHOOK);
	  _mutt_save_message(Context->hdrs[Context->v2r[i]],
			     &ctx, delete, decode, decrypt);
	}
      }
    }

    need_buffy_cleanup = (ctx.magic == M_MBOX || ctx.magic == M_MMDF || ctx.magic == M_KENDRA);

    mx_close_mailbox (&ctx, NULL);

    if (need_buffy_cleanup)
    {
#ifdef BUFFY_SIZE
      tmp = mutt_find_mailbox (buf);
      if (tmp && !tmp->new)
	mutt_update_mailbox (tmp);
#else
      /* fix up the times so buffy won't get confused */
      if (st.st_mtime > st.st_atime)
      {
	ut.actime = st.st_atime;
	ut.modtime = time (NULL);
	utime (buf, &ut); 
      }
      else
	utime (buf, NULL);
#endif
    }

    mutt_clear_error ();
    return (0);
  }
  
  return -1;
}


void mutt_version (void)
{
  mutt_message ("Mutt %s (%s)", MUTT_VERSION, ReleaseDate);
}

void mutt_edit_content_type (HEADER *h, BODY *b, FILE *fp)
{
  char buf[LONG_STRING];
  char obuf[LONG_STRING];
  char tmp[STRING];
  PARAMETER *p;

  char charset[STRING];
  char *cp;

  short charset_changed = 0;
  short type_changed = 0;
  
  cp = mutt_get_parameter ("charset", b->parameter);
  strfcpy (charset, NONULL (cp), sizeof (charset));

  snprintf (buf, sizeof (buf), "%s/%s", TYPE (b), b->subtype);
  strfcpy (obuf, buf, sizeof (obuf));
  if (b->parameter)
  {
    size_t l;
    
    for (p = b->parameter; p; p = p->next)
    {
      l = strlen (buf);

      rfc822_cat (tmp, sizeof (tmp), p->value, MimeSpecials);
      snprintf (buf + l, sizeof (buf) - l, "; %s=%s", p->attribute, tmp);
    }
  }
  
  if (mutt_get_field ("Content-Type: ", buf, sizeof (buf), 0) != 0 ||
      buf[0] == 0)
    return;
  
  /* clean up previous junk */
  mutt_free_parameter (&b->parameter);
  FREE (&b->subtype);
  
  mutt_parse_content_type (buf, b);

  
  snprintf (tmp, sizeof (tmp), "%s/%s", TYPE (b), NONULL (b->subtype));
  type_changed = ascii_strcasecmp (tmp, obuf);
  charset_changed = ascii_strcasecmp (charset, mutt_get_parameter ("charset", b->parameter));

  /* if in send mode, check for conversion - current setting is default. */

  if (!h && b->type == TYPETEXT && charset_changed)
  {
    snprintf (tmp, sizeof (tmp), _("Convert to %s upon sending?"),
	      mutt_get_parameter ("charset", b->parameter));
    b->noconv = !mutt_yesorno (tmp, !b->noconv);
  }

  /* inform the user */
  
  if (type_changed)
    mutt_message (_("Content-Type changed to %s."), tmp);
  else if (b->type == TYPETEXT && charset_changed)
    mutt_message (_("Character set changed to %s; %s."),
		  mutt_get_parameter ("charset", b->parameter),
		  b->noconv ? _("not converting") : _("converting"));

  b->force_charset |= charset_changed ? 1 : 0;

  if (!is_multipart (b) && b->parts)
    mutt_free_body (&b->parts);
  if (!mutt_is_message_type (b->type, b->subtype) && b->hdr)
  {
    b->hdr->content = NULL;
    mutt_free_header (&b->hdr);
  }

  if (fp && (is_multipart (b) || mutt_is_message_type (b->type, b->subtype)))
    mutt_parse_part (fp, b);
  
#if defined(HAVE_PGP) || defined(HAVE_SMIME) || defined(HAVE_GPGME)
  if (h)
  {
    if (h->content == b)
      h->security  = 0;

    h->security |= crypt_query (b);
  }
#endif
}


#if defined(HAVE_PGP) || defined(HAVE_GPGME)
static int _mutt_check_traditional_pgp (HEADER *h, int *redraw)
{
  MESSAGE *msg;
  int rv = 0;
  
  mutt_parse_mime_message (Context, h);
  if ((msg = mx_open_message (Context, h->msgno)) == NULL)
    return 0;
  if (pgp_check_traditional (msg->fp, h->content, 0))
  {
    h->security = crypt_query (h->content);
    *redraw |= REDRAW_FULL;
    rv = 1;
  }
  
  mx_close_message (&msg);
  return rv;
}

int mutt_check_traditional_pgp (HEADER *h, int *redraw)
{
  int i;
  int rv = 0;
  if (h)
    rv = _mutt_check_traditional_pgp (h, redraw);
  else
  {
    for (i = 0; i < Context->vcount; i++)
      if (Context->hdrs[Context->v2r[i]]->tagged)
	rv = _mutt_check_traditional_pgp (Context->hdrs[Context->v2r[i]], redraw)
	  || rv;
  }
  return rv;
}

#endif
